import edu.isi.integration.BSL.*; 
import edu.isi.integration.Phoebus.*;
import edu.isi.integration.Commons.*;
 
import com.wcohen.secondstring.*;

import java.util.*;
import java.sql.*;
import java.io.*;

public class RecordMatcher {
    private static final int RANK_K = 1; //can keep top K matches
    private static double THRESH = 0.95; //threshold for fuzzy token matching
    private static JaroWinklerTFIDF tfidf; //static object for doing tfidf similarity
    private static String whichType = "dice"; //can select dice or jaccard similarity, if you don't want to do tfidf
    private static boolean oneToOne = true; //can enforce 1-1 matching, or 1-N matching

    public static void setThresh(double d) {
	THRESH = d;
    }

    public static void main(String args[]) {
	oneToOne = false;
	String postSetFile = args[0];
	String postSetName = args[1];
	String refSetFile = args[2];
	String refSetName = args[3];
	double dd  = (new Double(args[4])).doubleValue();
	setThresh(dd);
	setType(args[5]);

	//now process all of them...
	XMLReader xr = new XMLReader();
	DataSet postSet = new DataSet();
	DataSet refSet = new DataSet();
		
	loadDataSet(postSetFile,postSetName,postSet,xr);
	loadDataSet(refSetFile,refSetName,refSet,xr);
	clearMatches(postSet);
	scoreDataSets(postSet, refSet);
    }

    public static void setType(String blah) {
	whichType = blah;
    }

    public static void scoreDataSets(DataSet postSet, DataSet refSet) {
	HashMap finalScores = new HashMap();

	//these are the strings that we'll use to hold each data set to treat them as 1 big document
	Vector stringWraps = new Vector();

	//iterate over refset set, grab full tuple as a string and make string wrappers for them to accum. stats
	//also add it to the big "document"
	refSet.setTableVals();
	HashMap refTuples = refSet.getTableVals();
	Vector refAttrs = refSet.getAttributes();
	Iterator refIter = refTuples.keySet().iterator();
	while(refIter.hasNext()) {
	    String currRefId =  (String)refIter.next();
	    HashMap attrs = (HashMap)refTuples.get((String)currRefId);
	    String full = "";
	    for(int i = 0; i < refAttrs.size(); i++) {
		full += (String)attrs.get((String)refAttrs.get(i))+" ";
	    }
	    full = full.trim();
	    //stringWraps.add(new StringWrapper(full.toLowerCase()));
	}

	//now, load up each string and compare it to teh members of the reference set...
  	postSet.setTableVals();
 	HashMap posts = postSet.getTableVals();

	if(whichType.equals("tfidf")) {
	    Iterator pit = posts.keySet().iterator();
	    while(pit.hasNext()) {
		String currId = (String)pit.next();
		HashMap atts = (HashMap)posts.get((String)currId);
		String pst = (String)atts.get("post");
		stringWraps.add(new StringWrapper(pst.toLowerCase()));
	    }
	    tfidf = new JaroWinklerTFIDF();
	    tfidf.setTokenMatchThreshold(THRESH);
	    tfidf.accumulateStatistics(stringWraps.iterator());
	}

 	Iterator postIter = posts.keySet().iterator();
	postSet.openConnection();
	Statement stmt = postSet.getStatement();
 	while(postIter.hasNext()) {
	    String currPostId =  (String)postIter.next();
	    HashMap attrs = (HashMap)posts.get((String)currPostId);
	    String post = (String)attrs.get("post");
	    Vector topK = scorePost(refTuples, refAttrs, post);
	    insertBestMatches(stmt, currPostId, topK, postSet, refSet);
	}
	postSet.closeConnection();
	
    }

    public static Vector scorePost(HashMap refTuples, Vector refAttrs, String post) {
	double topScores[] = new double[RANK_K];
	Vector topRefs[] = new Vector[RANK_K];

	//Jaccard jacc = new Jaccard();
	//JaroWinkler jw = new JaroWinkler();
	Iterator refIt = refTuples.keySet().iterator();
	for(int i = 0; i < RANK_K; i++) {
	    topScores[i] = -1.0;
	    topRefs[i] = new Vector();
	}
	while(refIt.hasNext()) {
	    String currRefId =  (String)refIt.next();
	    HashMap attrs = (HashMap)refTuples.get((String)currRefId);
	    String full = "";
	    for(int i = 0; i < refAttrs.size(); i++) {
		full += (String)attrs.get((String)refAttrs.get(i))+" ";
	    }
	    full = full.trim();

	    full = full.replaceAll("\\s+"," ");
	    post = post.replaceAll("\\s+"," ");
	    double score = 0.0;
	    if(whichType.equals("tfidf")) {
		score = tfidf.score(tfidf.prepare(full.toLowerCase()), tfidf.prepare(post.toLowerCase()));
	    }
	    else {
		score = getJWDice(full, post);
	    }

	    //Update the set of scores
	    int maxIndex = -1;
	    int eqIndex = -1;
	    for(int j = 0; j < RANK_K; j++) {
		if(score >= topScores[j]) {
		    if(score > topScores[j]) {
			maxIndex = j;
			break;
		    }
		    if(score == topScores[j]) {
			eqIndex = j;
			break;
		    }
		}
	    }

	    //now max score should point to the index where the hightest score is
	    //also, eq index points to index where it is the same
	    if(eqIndex > -1) {
		topRefs = addTopRef(topRefs, (new Integer(currRefId)).intValue(),eqIndex);
	    }
	    if(maxIndex > -1) {
		topScores = updateTopScores(topScores, score, maxIndex);
		topRefs = updateTopRefs(topRefs, (new Integer(currRefId)).intValue(), maxIndex);
	    }
	}
	System.out.println("FOR POST: "+post);
	Vector retVec = new Vector();
	
	for(int a = 0; a < RANK_K; a++) {
	    Vector intRefs = topRefs[a];
	    if(oneToOne) {
		if(intRefs.size() == 1) {
		    for(int b = 0; b < intRefs.size(); b++) {
			Integer intV = new Integer((String)intRefs.get(b));
			String refId = (String)intV.toString();
			HashMap attrs = (HashMap)refTuples.get((String)refId);
			String full = "";
			for(int i = 0; i < refAttrs.size(); i++) {
			    full += (String)attrs.get((String)refAttrs.get(i))+" ";
			}
			full = full.trim();
			double sc = topScores[a];
			String strD = (String)(new Double(sc)).toString();
			System.out.println("RANK "+(a+1)+": "+full+" -- "+sc);
			retVec.add((String)refId+"|"+((String)strD)+"|"+(a+1));
		    }
		}
	    }
	    else {
		for(int b = 0; b < intRefs.size(); b++) {
		    Integer intV = new Integer((String)intRefs.get(b));
		    String refId = (String)intV.toString();
		    HashMap attrs = (HashMap)refTuples.get((String)refId);
		    String full = "";
		    for(int i = 0; i < refAttrs.size(); i++) {
			full += (String)attrs.get((String)refAttrs.get(i))+" ";
		    }
		    full = full.trim();
		    double sc = topScores[a];
		    String strD = (String)(new Double(sc)).toString();
		    System.out.println("RANK "+(a+1)+": "+full+" -- "+sc);
		    retVec.add((String)refId+"|"+((String)strD)+"|"+(a+1));
		}
	    }
	}
	System.out.println("----------------");	    
	return retVec;	
    }

    public static Vector[] addTopRef(Vector topRefs[], int ref, int eqIndex) {
	Vector retVects[] = new Vector[RANK_K];
	for(int i = 0; i < RANK_K; i++) {
	    if(i == eqIndex) {
		Vector eqIndV = (Vector)topRefs[i];
		String cr = (new Integer(ref)).toString();
		eqIndV.add((String)cr);
		retVects[i] = (Vector)eqIndV;
	    }
	    else {
		retVects[i] = topRefs[i];
	    }
	}
	return retVects;
    }

    public static double[] updateTopScores(double topScores[], double score, int maxIndex) {
	double retVal[] = new double[RANK_K];
	for(int i = 0; i < RANK_K; i++) {
	    if(i < maxIndex) {
		retVal[i] = topScores[i];
	    }
	    else {
		if( i == maxIndex) {
		    retVal[i] = score;
		}
		else { //i > maxIndex
		    retVal[i] = topScores[i-1];
		}
	    }
	}
	return retVal;
    }

    public static Vector[] updateTopRefs(Vector topScores[], int ref, int maxIndex) {
	Vector retVect[] = new Vector[RANK_K];
	for(int i = 0; i < RANK_K; i++) {
	    if(i < maxIndex) {
		retVect[i] = topScores[i];
	    }
	    else {
		if( i == maxIndex) {
		    Vector v = new Vector();
		    String cr = (new Integer(ref)).toString();
		    v.add((String)cr);
		    retVect[i] = v;
		}
		else { //i > maxIndex
		    retVect[i] = topScores[i-1];
		}
	    }
	}
	return retVect;
    }

    public static void loadDataSet(String file, String setName, DataSet ds, XMLReader xr) {
	xr.retrieveDataSet(file, setName, ds);
    }

    public static void clearMatches(DataSet postSet) {
	try {
	    postSet.openConnection();
	    String q = "DELETE FROM irMatches";
	    Statement st = (postSet.getStatement());
	    st.executeUpdate(q);
	    postSet.closeConnection();
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
    }

    public static void insertBestMatches(Statement stmt, String postId, Vector topK, DataSet postSet, DataSet refSet) {
	String q = "";
	try {
	    for(int i = 0; i < topK.size(); i++) {
		String s[] = ((String)topK.get(i)).split("\\|");	    
		String id = s[0];
		double sc = (new Double(s[1])).doubleValue();
		int rank = (new Integer(s[2])).intValue();
		String pidName = postSet.getTableId();
		String ridName = refSet.getTableId();
		if(sc > 0.0) {
		    q = "INSERT INTO irMatches("+pidName+", "+ridName+",rank,score) VALUES('"+postId+"','"+id+"',"+rank+","+sc+")";
		    stmt.executeUpdate(q);
		}
	    }
	}
	catch(Exception e) {
	    System.out.println("FAILED: "+q);
	    e.printStackTrace();
	}
    }

    /***** GET THE Jaro-Winkler Dice Sim Score*/
    public static double getJWDice(String refStr, String postStr) {
	//double THRESH = 1.0;
 	refStr = refStr.toLowerCase();
	postStr = postStr.toLowerCase();
	String refToks[] = refStr.split(" ");
	String postToks[] = postStr.split(" ");
	int intersection = 0;
	TreeSet usedPostToks = new TreeSet();
	JaroWinkler jw = new JaroWinkler();
	for(int r = 0; r < refToks.length; r++) {
	    for(int t = 0; t < postToks.length; t++) {
		double sc = jw.score(jw.prepare(refToks[r].toLowerCase()), jw.prepare(postToks[t].toLowerCase()));
		//double sc = jw.getSimilarity(refToks[r],postToks[t]);
		String ti = (String)(new Integer(t)).toString();
		if(sc >= THRESH  && !usedPostToks.contains((String)ti)) {
		//if(refToks[r].equalsIgnoreCase(postToks[t]) && !usedPostToks.contains((String)ti)) {
		    //System.out.println("TOK MATCH: "+refToks[r]+" VS "+postToks[t]);
		    intersection++;
		    usedPostToks.add((String)ti);
		}
	    }
	}	

	double sc = 0.0;
	if(whichType.equals("dice")) {
	    //Dices coefficient = (2*Common Terms) / (Number of terms in String1 + Number of terms in String2)
	    sc = ((2.0*(double)intersection) / (double)((double)refToks.length + (double)postToks.length));
	}
	else {
	    //Jaccard similarity: Note that the union is the same as the size of the sets minus the intersection
	    sc = (double)intersection / ((double)refToks.length + (double)postToks.length - (double)intersection);
	    //System.out.println(refStr+" VS "+postStr+" EQ "+sc);
	}
	return sc;
    }
    //*******************************************/
}