import edu.isi.integration.BSL.*; 
import edu.isi.integration.Phoebus.*;
import edu.isi.integration.Commons.*;

import java.util.*;
import java.sql.*;
import java.io.*;

public class Experiment {
    private Vector blockingScheme;

    public static void main(String args[]) {
	Experiment e = new Experiment();
	if(args[0].equals("run")) {
	    e.train(args[1],args[2],args[3],args[4],args[5],args[6],args[7],args[8],args[9].split(","),args[13],args[15]);
	    e.test(args[10],args[11],args[12],args[4],args[5],args[6],args[8],args[9].split(","),args[2],args[3],args[7],args[15]);
	    e.getResults(args[11],args[12],args[13],args[5],args[6],args[14],args[4]); 
	}
    }
    
    public void test(String postIdFile, String postDataSetFile, 
		      String postDataSetName, String allPostTableName,
		      String refSetDataSetFile, String refSetDataSetName,
		     String commonAttributeXMLFile, String chosenCommonScores[],
		     String trainingPostDataSetFile, String trainingPostDataSetName,
		     String attrMethodFile, String isOneToOne) {
	//isOneToOne?
	boolean isOne2One = false;
	if(isOneToOne.equals("true")) {
	    isOne2One = true;
	}
	
	//we need to make the datasets
	XMLReader xr = new XMLReader();
	DataSet postSet = new DataSet();
	DataSet refSet = new DataSet();
	DataSet trainingSet = new DataSet();//just used for the name...
 	loadDataSet(postDataSetFile,postDataSetName,postSet,xr);
	loadDataSet(refSetDataSetFile,refSetDataSetName,refSet,xr);
	loadDataSet(trainingPostDataSetFile,trainingPostDataSetName,trainingSet,xr);

	DataSet matchTable = new DataSet();
	matchTable.setName("matches");
	matchTable.setTableName(postSet.getTableName()+"_"+refSet.getTableName()+"_MATCHES");
	matchTable.setURL(postSet.getURL());
	matchTable.setLogin(postSet.getLogin());
	matchTable.setPass(postSet.getPass());
	matchTable.addAttribute(postSet.getTableId());
	matchTable.addAttribute(refSet.getTableId());
	clearMatchTable(matchTable);


	//load up the commonScores
	HashMap commonScores = new HashMap();
	commonScores = xr.retrieveCommonScores(commonAttributeXMLFile, chosenCommonScores);

	//first we load up the training data
	HashMap testPosts = getPostsFromFile(postIdFile, postDataSetFile, postDataSetName, allPostTableName);
	TableMaker tm = new TableMaker();
	tm.fillTableWithPosts(testPosts, postSet, matchTable.getTableName(), refSet.getTableId());
	//update the attributes to be nice blank strings, instead of nulls
	Vector refSchema = refSet.getAttributes();
	for(int zz = 0; zz < refSchema.size(); zz++) {
	    tm.updateField((String)refSchema.get(zz),"",postSet);
	}
	Set cScoreSet = commonScores.keySet();
	Vector cScores = new Vector((Collection)cScoreSet);
	for(int aa = 0; aa < cScores.size(); aa++) {
	    tm.updateField((String)cScores.get(aa),"",postSet);
	}


	//set it up for blocking
	BSL bsl = new BSL();
	//bsl.setPath(dirPath);
	//set up for blocking...(copy post to all attributes)
	bsl.setUpTable(postSet,"blocking");
	//now we generate the candidates using the blocking scheme
	//System.out.println("Generating candidates with blocking scheme...At:"+new Time(System.currentTimeMillis()));
	long startTime = System.currentTimeMillis();
	bsl.runBlockingScheme(blockingScheme,postSet,refSet,true); 
	long stopTime = System.currentTimeMillis();
	System.out.println("BLOCKING SCHEME RUN IN: "+(stopTime - startTime)+" MILLIS");

	//bsl.setUpTable(postSet, blockingScheme, "phoebus");
	bsl.setUpTable(postSet,"phoebus");

	//now we run Phoebus
	Phoebus p = new Phoebus();
	p.setOneToOne(isOne2One);
 	p.setMode("test");
	p.setPath("./");
	p.setCandidateFile(postSet.getTableName()+"_finalCands.txt");
	p.setClassFile(postSet.getTableName()+"Class.dat");
	//System.out.println("Build classification feature vectors...At:"+new Time(System.currentTimeMillis()));
	startTime = System.currentTimeMillis();
	p.buildSVMClassData(postSet, refSet, matchTable);
	p.runSVMClassifier(postSet.getTableName()+"Class.dat",trainingSet.getTableName()+"ClassModel",postSet.getTableName()+"ClassPreds.txt");
	stopTime = System.currentTimeMillis();
	System.out.println("RAN CLASSIFIER IN: "+(stopTime - startTime)+" MILLIS");
	//System.out.println("Get matches...At:"+new Time(System.currentTimeMillis()));
	p.getMatches(postSet.getTableName()+"Class.dat",postSet.getTableName()+"ClassPreds.txt",refSet,postSet,matchTable);
	p.setSegFile(postSet.getTableName()+"TestSeg.dat");
	System.out.println("Build extraction feature vectors...At:"+new Time(System.currentTimeMillis()));
	p.buildSVMSegmentationData(postSet, refSet, matchTable, commonScores);
	p.runSVMSegmenter(postSet.getTableName()+"TestSeg.dat",trainingSet.getTableName()+"SegModel",postSet.getTableName()+"SegPreds.txt");
	System.out.println("Label and post process the extracted tokens...At:"+new Time(System.currentTimeMillis()));
	p.getLabeledTokens(postSet.getTableName()+"TestSeg.dat",postSet.getTableName()+"SegPreds.txt",postSet,commonScores);  
	p.postProcessFields(postSet, refSet, matchTable);

    }

    public void train(String postIdFile, String postDataSetFile, 
		      String postDataSetName, String allPostTableName,
		      String refSetDataSetFile, String refSetDataSetName,
		      String attrMethodFile, String commonAttributeXMLFile,
		      String chosenCommonScores[], String trueMatchesTableName,
		      String isOneToOne) {
	//isOneToOne?
	boolean isOne2One = false;
	if(isOneToOne.equals("true")) {
	    isOne2One = true;
	}

	//first we load up the training data
	readAndPut(postIdFile, postDataSetFile, postDataSetName, allPostTableName);
	
	//we need to make the datasets
	XMLReader xr = new XMLReader();
	DataSet postSet = new DataSet();
	DataSet refSet = new DataSet();
	loadDataSet(postDataSetFile,postDataSetName,postSet,xr);
	loadDataSet(refSetDataSetFile,refSetDataSetName,refSet,xr);


	DataSet matchTable = new DataSet();
	matchTable.setName("matches");
	matchTable.setTableName(postSet.getTableName()+"_"+refSet.getTableName()+"_MATCHES");
	matchTable.setURL(postSet.getURL());
	matchTable.setLogin(postSet.getLogin());
	matchTable.setPass(postSet.getPass());
	matchTable.addAttribute(postSet.getTableId());
	matchTable.addAttribute(refSet.getTableId());

	loadMatchTable(postSet,refSet,matchTable,trueMatchesTableName);

	//next we run the blocker
	//set it up for blocking
	BSL bsl = new BSL();
	//bsl.setPath(dirPath);
	//set up for blocking...(copy post to all attributes)
	bsl.setUpTable(postSet,"blocking");
	//run blocker
	long startTime = System.currentTimeMillis();
	blockingScheme = bsl.learnRule(postSet,refSet,matchTable,attrMethodFile,true);
	long stopTime = System.currentTimeMillis();
	System.out.println("BLOCKING RULE LEARNED IN: "+(stopTime - startTime)+" MILLIS");

	//now we reload the data again, since we overwrote the attributes for the blocking stuff...
	readAndPut(postIdFile, postDataSetFile, postDataSetName, allPostTableName);

	//load up the commonScores
	HashMap commonScores = new HashMap();
	commonScores = xr.retrieveCommonScores(commonAttributeXMLFile, chosenCommonScores);

	//now we run phoebus train and leanrn a model..
	Phoebus p = new Phoebus();
	p.setOneToOne(isOne2One);
	if(refSetDataSetName.equals("cars")) {
	    p.setMode("train_large");
	}
	else {
	    p.setMode("train");
	}
	p.setPath("./");
	p.setCandidateFile(postSet.getTableName()+"_finalCands.txt");
	p.setClassFile(postSet.getTableName()+"Class.dat");
	p.buildSVMClassData(postSet, refSet, matchTable);
	p.trainSVMClassifier(postSet.getTableName()+"Class.dat",postSet.getTableName()+"ClassModel");
	p.setSegFile(postSet.getTableName()+"TrainSeg.dat");
	p.buildSVMSegmentationData(postSet, refSet, matchTable, commonScores);
	p.trainSVMSegmenter(postSet.getTableName()+"TrainSeg.dat",postSet.getTableName()+"SegModel");
    }


    public void loadDataSet(String file, String setName, DataSet ds, XMLReader xr) {
	xr.retrieveDataSet(file, setName, ds);
	//System.out.println(ds);
    }

    public static void clearAttributes(DataSet postSet, DataSet refSet) {
	try {
	    Vector attributes = postSet.getAttributes();
	    postSet.openConnection();
	    for(int i = 0; i < attributes.size(); i++) {
		String attr = (String)attributes.get(i);
		if(!(attr.equals("post") || attr.equals(postSet.getTableId()) || attr.equals(refSet.getTableId()))) {
		    String q = "UPDATE "+postSet.getTableName()+" SET "+attr+"=''";
		    (postSet.getStatement()).executeUpdate(q);
		}
		if(attr.equals(refSet.getTableId())) {
		    String q = "UPDATE "+postSet.getTableName()+" SET "+attr+"='0'";
		    (postSet.getStatement()).executeUpdate(q);
		}
	    }
	    postSet.closeConnection();
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
    }

    public static void readAndPutPost(String idFile, String dataSetFile, String dataSetName, String fullTableName) {
	try {
	    XMLReader xr = new XMLReader();
	    DataSet ds = new DataSet();
	    xr.retrieveDataSet(dataSetFile,dataSetName,ds);
	    BufferedReader in = new BufferedReader(new FileReader(idFile));
	    String str = ""; 
	    ds.openConnection();
	    String q = "DELETE FROM "+ds.getTableName();
	    (ds.getStatement()).executeUpdate(q);
	    while ((str = in.readLine()) != null) {
		q = "INSERT INTO "+ds.getTableName()+" (post) SELECT post FROM "+fullTableName+" WHERE "+ds.getTableId()+"='"+str+"'";
		(ds.getStatement()).executeUpdate(q);
	    } //while
	    ds.closeConnection();
	} //try
	catch(Exception e) { e.printStackTrace();}
    }

    public static HashMap getPostsFromFile(String idFile, String dataSetFile, String dataSetName,String fullTableName) {
	HashMap ret = new HashMap();
	try {
	    XMLReader xr = new XMLReader();
	    DataSet ds = new DataSet();
	    xr.retrieveDataSet(dataSetFile,dataSetName,ds);
	    BufferedReader in = new BufferedReader(new FileReader(idFile));
	    String str = ""; 
	    ds.openConnection();
	    String q = "DELETE FROM "+ds.getTableName();
	    (ds.getStatement()).executeUpdate(q);
	    while ((str = in.readLine()) != null) {
		q = "SELECT post FROM "+fullTableName+" WHERE "+ds.getTableId()+"='"+str+"'";
		ResultSet rs = (ds.getStatement()).executeQuery(q);
		while(rs.next()) {
		    String post = rs.getString("post");
		    ret.put((String)str,(String)post);
		}
	    } //while
	    ds.closeConnection();
	} //try
	catch(Exception e) { e.printStackTrace();}
	return ret;
    }

    public void getResults(String postDataSetFile, 
			   String postDataSetName, String truthTableName,
			   String refSetDataSetFile, String refSetDataSetName,
			   String resultsFile, String origPostsTable) {
	XMLReader xr = new XMLReader();
	DataSet postSet = new DataSet();
	DataSet refSet = new DataSet();
	loadDataSet(postDataSetFile,postDataSetName,postSet,xr);
	loadDataSet(refSetDataSetFile,refSetDataSetName,refSet,xr);

	DataSet matchTable = new DataSet();
	matchTable.setName("matches");
	matchTable.setTableName(postSet.getTableName()+"_"+refSet.getTableName()+"_MATCHES");
	matchTable.setURL(postSet.getURL());
	matchTable.setLogin(postSet.getLogin());
	matchTable.setPass(postSet.getPass());
	matchTable.addAttribute(postSet.getTableId());
	matchTable.addAttribute(refSet.getTableId());
	
	String esTok = extractionStats(postSet,origPostsTable,refSet,matchTable,truthTableName,"token");
	String esField = extractionStats(postSet,origPostsTable,refSet,matchTable,truthTableName,"field");
	String blockRes = blockingStats(postSet,refSet,matchTable,truthTableName);
	String recLinkRes = recordLinkageStats(postSet,refSet,matchTable,truthTableName);
	try {
	    BufferedWriter out = new BufferedWriter(new FileWriter(resultsFile,true));
	    out.write(blockRes);
	    out.write("------------\n");
	    out.write(recLinkRes);
	    out.write("------------\n");
	    out.write(esTok);
	    out.write("------------\n");
	    out.write(esField);
	    out.write("##############");
	    out.newLine();
	    out.close();
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
    }

    public String extractionStats(DataSet postSet, String origPostsTable, DataSet refSet, DataSet matchTable, String truthTableName, String method) {
	//we use origPostsTable to see which extractions are correct, and truthTableName for the true matches...
	String ret = "";
	Vector attrs = postSet.getAttributes();
	attrs.remove((String)"post");
	int possCounts[] = new int[attrs.size()];
	int madeCounts[] = new int[attrs.size()];
	int matchCounts[] = new int[attrs.size()];
	for(int z = 0; z < possCounts.length; z++) {
	    possCounts[z] = 0;
	    madeCounts[z] = 0;
	    matchCounts[z] = 0;
	}

	//to count the correct extractions that we did and the total possible correct ones
	String q = "SELECT DISTINCT ";
	q += "p."+postSet.getTableId()+",";
	for(int i = 0; i < attrs.size(); i++) {
	    q += "o."+(String)attrs.get(i)+" AS truth"+(String)attrs.get(i)+",";
	    q += "p."+(String)attrs.get(i)+" AS made"+(String)attrs.get(i)+",";
	}
	q += " FROM ";
	q = q.replaceAll(", FROM"," FROM");
	q += postSet.getTableName()+" p,"+origPostsTable+" o,"+matchTable.getTableName()+" m ";
	if((refSet.getTableName()).equalsIgnoreCase("cars")) {
	    q += ", CarsWithExtractedValues cwev ";
	}
	q += " WHERE "; 
	q += "p."+postSet.getTableId()+"=o."+postSet.getTableId();
	q += " AND m."+refSet.getTableId()+" NOT LIKE '0' AND p."+postSet.getTableId()+"=m."+postSet.getTableId();
	if((refSet.getTableName()).equalsIgnoreCase("cars")) {
	    q += " AND p."+postSet.getTableId()+"=cwev."+postSet.getTableId();
	}
	//q += " AND o."+postSet.getTableId()+" IN (SELECT "+postSet.getTableId()+" FROM "+truthTableName+")";
	//System.out.println(q);
	try {
	    postSet.openConnection();
	    ResultSet rs = (postSet.getStatement()).executeQuery(q);
	    while(rs.next()) {
		for(int a = 0; a < attrs.size(); a++) {
		    String sTruth = rs.getString("truth"+(String)attrs.get(a));
		    String sMade = rs.getString("made"+(String)attrs.get(a));
		    if(sMade != null && sMade.trim().length() > 0) {
			String s = sMade;
			if(method.equals("token")) {
			    s = s.trim();
			    s = s.replaceAll("\\s+"," ");
			    String toks[] = s.split(" ");
			    //System.out.println("String: "+s+" HAS "+toks.length+" TOKENS");
			    madeCounts[a] += toks.length;
			}
			if(method.equals("field")) {
			    madeCounts[a]++;
			}
		    }
		    if(sTruth != null && sTruth.trim().length() > 0) {
			String s = sTruth;
			if(method.equals("token")) {
			    s = s.replaceAll("\\s+"," ");
			    String toks[] = s.split(" ");
			    //System.out.println("String: "+s+" HAS "+toks.length+" TOKENS");
			    possCounts[a] += toks.length;
			}
			if(method.equals("field")) {
			    possCounts[a]++;
			}
		    }
		    if(sTruth != null && sMade != null && sTruth.trim().length() > 0
		       && sMade.trim().length() > 0) {
			if(method.equals("token")) {
			    sTruth = sTruth.replaceAll("\\s+"," ");
			    String toksTruth[] = sTruth.split(" ");

			    sMade = sMade.replaceAll("\\s+"," ");
			    String toksMade[] = sMade.split(" ");
			    for(int x = 0; x < toksTruth.length; x++) {
				for(int y = 0; y < toksMade.length; y++) {
				    if(toksTruth[x].trim().length() > 0 && 
				       toksMade[y].trim().length() > 0 && 
				       toksTruth[x].equals(toksMade[y])) {
					matchCounts[a]++;					
					toksTruth[x] = "";
					toksMade[y] = "";
				    }
				}
			    }
			    //System.out.println("String: "+s+" HAS "+toks.length+" TOKENS");			    
			}
			if(method.equals("field")) {
			    sTruth = sTruth.replaceAll("\\s+"," ");
			    String toksTruth[] = sTruth.split(" ");

			    sMade = sMade.replaceAll("\\s+"," ");
			    String toksMade[] = sMade.split(" ");
			    if(toksTruth.length == toksMade.length) {
				for(int x = 0; x < toksTruth.length; x++) {
				    for(int y = 0; y < toksMade.length; y++) {
					if(toksTruth[x].trim().length() > 0 && toksMade[y].trim().length() > 0 && 
					   toksTruth[x].equals(toksMade[y])) {
					    toksTruth[x] = "";
					    toksMade[y] = "";
					}
				    }
				}
				int found = 0;
				for(int zz = 0; zz < toksTruth.length; zz++) {
				    if(toksTruth[zz].trim().length() > 0 ||
				       toksMade[zz].trim().length() > 0) {
					found = 1;
				    }
				}
				if(found == 0) {
				    matchCounts[a]++;
				}
			    }
			}
		    }
		}
	    }
	    postSet.closeConnection();
	}
	catch(Exception e) {
	    e.printStackTrace();
	}

	for(int b = 0; b < attrs.size(); b++) {
	    double recall = (double)matchCounts[b] / (double)possCounts[b];
	    double precision = (double)matchCounts[b] / (double)madeCounts[b];
	    System.out.println("ATTR: "+(String)attrs.get(b)+" POSS:"+possCounts[b]+" MADE:"+madeCounts[b]+" CORR:"+matchCounts[b]+" RECALL:"+recall+" PREC:"+precision);
	    ret += method.toUpperCase()+"_ATTR_"+(String)attrs.get(b)+"_RECALL:"+recall+"\n";
	    ret += method.toUpperCase()+"_ATTR_"+(String)attrs.get(b)+"_PRECISION:"+precision+"\n";
	}

	return ret;
    }

    public String blockingStats(DataSet postSet, DataSet refSet, 
				     DataSet matchTable, String truthTableName) {
	int numCands = 0;
	int numCorrCands = 0;
	String ret = "";
	try {
	    BufferedReader in = new BufferedReader(new FileReader(postSet.getTableName()+"_finalCands.txt"));
	    String str = ""; 
	    postSet.openConnection();
	    Statement st = postSet.getStatement();
	    while ((str = in.readLine()) != null) {
		String parts[] = str.split("\\|");
		String postId = parts[0];
		String refId = parts[1];
		numCands++;
		if(isTrueMatch(postId,refId,postSet.getTableId(),refSet.getTableId(),truthTableName,st)) {
		    numCorrCands++;
		}
	    } //while
	    double crossProdSize = getCrossProd(postSet.getTableName(),refSet.getTableName(),st);
	    postSet.closeConnection();
	    int numPossCorr = getNumPossCorr(postSet,refSet,matchTable,truthTableName);
	    ret += "************\n";
	    ret += "BS: "+blockingScheme+"\n";
	    ret += "NUM CORR:\t"+numCorrCands+"\n";
	    ret += "NUM CANDS:\t"+numCands+"\n";
	    ret += "NUM POSS:\t"+numPossCorr+"\n";
	    ret += "CROSS PROD:\t"+crossProdSize+"\n";
	    double reductionRatio = 1.0 - (double)numCands / crossProdSize;
	    double pairsCompleteness = (double)numCorrCands / (double)numPossCorr;
	    ret += "RR:\t"+reductionRatio+"\n";
	    ret += "PC:\t"+pairsCompleteness+"\n";
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
	return ret;
    }

    public static int getNumPossCorr(DataSet postSet, DataSet refSet, 
				      DataSet matchTable, String truthTableName) {
	int numPoss = 0;
	try {
	    postSet.openConnection();
	    //Num poss:
	    String q = "SELECT m.* FROM "+postSet.getTableName()+" p, "+truthTableName+" m WHERE p."+postSet.getTableId()+"=m."+postSet.getTableId();
	    ResultSet rs = (postSet.getStatement()).executeQuery(q);
	    while(rs.next()) {
		numPoss++;
	    }
	    postSet.closeConnection();
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
	return numPoss;
    }

    public static int getNumCorr(DataSet postSet, DataSet refSet, 
				 DataSet matchTable, String truthTableName) {
	int numCorr = 0;
	String pId = postSet.getTableId();
	String rId = refSet.getTableId();
	try {
	    postSet.openConnection();
	    //num corr query
	    String q = "SELECT DISTINCT  m."+pId+",m."+rId+" FROM "+matchTable.getTableName()+" p, "+truthTableName+" m WHERE p."+pId+"=m."+pId+" AND p."+rId+"=m."+rId;
	    ResultSet rs = (postSet.getStatement()).executeQuery(q);
	    while(rs.next()) {
		numCorr++;
	    }
	    postSet.closeConnection();
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
	return numCorr;
    }

    public static int getNumMade(DataSet postSet, DataSet refSet, 
				 DataSet matchTable, String truthTableName) {
	int numMade = 0;
	String pId = postSet.getTableId();
	String rId = refSet.getTableId();
	try {
	    postSet.openConnection();
	    //num corr query
	    String q = "SELECT distinct "+postSet.getTableId()+","+refSet.getTableId()+" FROM "+matchTable.getTableName()+" WHERE "+refSet.getTableId()+" NOT LIKE '0'";
	    ResultSet rs = (postSet.getStatement()).executeQuery(q);
	    while(rs.next()) {
		numMade++;
	    }
	    postSet.closeConnection();
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
	return numMade;
    }

    public static double getCrossProd(String postSetTableName, String refSetTableName, Statement st) {
	double postSetSize = 0.0;
	double refSetSize = 0.0;
	try {
	    String q = "SELECT COUNT(*) AS C FROM "+postSetTableName;
	    ResultSet rs = st.executeQuery(q);
	    while(rs.next()) {
		postSetSize = 1.0*(new Integer(rs.getString("C"))).intValue();
	    }

	    q = "SELECT COUNT(*) AS C FROM "+refSetTableName;
	    rs = st.executeQuery(q);
	    while(rs.next()) {
		refSetSize = 1.0*(new Integer(rs.getString("C"))).intValue();
	    }
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
	return postSetSize*refSetSize;
    }
    public static boolean isTrueMatch(String postId, String refId,
				      String postIdAttr, String refIdAttr,
				      String matchTableName, Statement st) {
	int ret = 0;
	String q = "";
	try {
	    q = "SELECT * FROM "+matchTableName+" WHERE "+postIdAttr+"='"+postId+"' AND "+refIdAttr+"='"+refId+"'";
	    ResultSet rs = st.executeQuery(q);
	    while(rs.next()) {
		ret++;
	    }
	    if(ret > 0) {
		return true;
	    }
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
	return false;
    }

    public static String recordLinkageStats(DataSet postSet, DataSet refSet, 
					  DataSet matchTable, String truthTableName) {
	String ret = "";
	int numCorr = getNumCorr(postSet,refSet,matchTable,truthTableName);
	int numMade = getNumMade(postSet,refSet,matchTable,truthTableName);
	int numPoss = getNumPossCorr(postSet,refSet,matchTable,truthTableName);
	try {
	    double recall = (double)numCorr/ (double)numPoss;
	    double prec = (double)numCorr / (double)numMade;
	    ret += "***************\n";
	    ret += "Num Corr:\t"+numCorr+"\n";
	    ret += "Num Made:\t"+numMade+"\n";
	    ret += "Num Poss:\t"+numPoss+"\n";
	    ret += "RLREC:\t"+recall+"\n";
	    ret += "RLPRECISION:\t"+prec+"\n";
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
	return ret;
    }

    public static void readAndPut(String idFile, String dataSetFile, String dataSetName,String fullTableName) {
    	try {
	    XMLReader xr = new XMLReader();
	    DataSet ds = new DataSet();
	    xr.retrieveDataSet(dataSetFile,dataSetName,ds);
	    BufferedReader in = new BufferedReader(new FileReader(idFile));
	    String str = ""; 
	    ds.openConnection();
	    String q = "DELETE FROM "+ds.getTableName();
	    (ds.getStatement()).executeUpdate(q);
	    while ((str = in.readLine()) != null) {
		q = "INSERT INTO "+ds.getTableName()+" SELECT * FROM "+fullTableName+" WHERE "+ds.getTableId()+"='"+str+"'";
		(ds.getStatement()).executeUpdate(q);
	    } //while
	    ds.closeConnection();
	} //try
	catch(Exception e) { e.printStackTrace();}
    }

    public static void clearMatchTable(DataSet matchTable) {
	String createStr = "CREATE TABLE IF NOT EXISTS "+matchTable.getTableName()+" (";
	String createCpyStr = "CREATE TABLE IF NOT EXISTS "+matchTable.getTableName()+"_COPY (";
	Vector atts = matchTable.getAttributes();
	for(int i = 0; i < atts.size(); i++) {
		if(i > 0) {
			createStr += ",";
			createCpyStr += ",";
		}
		createStr += (String)atts.get(i)+" text";
		createCpyStr += (String)atts.get(i)+" text";
	}
	createStr += ")";
	createCpyStr += ")";

	String del = "DELETE FROM "+matchTable.getTableName();
	String delCpy = "DELETE FROM "+matchTable.getTableName()+"_COPY";
	try {
	    matchTable.openConnection();
	    (matchTable.getStatement()).executeUpdate(createStr);
	    (matchTable.getStatement()).executeUpdate(createCpyStr);
	    (matchTable.getStatement()).executeUpdate(del);
	    (matchTable.getStatement()).executeUpdate(delCpy);
	    matchTable.closeConnection();
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
    }
    public static void loadMatchTable(DataSet postSet,DataSet refSet,DataSet matchTable,String trueMatchesTableName) {
	String q = "INSERT INTO "+matchTable.getTableName()+" ("+postSet.getTableId()+","+refSet.getTableId()+")";
	q += " SELECT tm."+postSet.getTableId()+",tm."+refSet.getTableId()+" FROM "+trueMatchesTableName+" tm, "+postSet.getTableName()+" p WHERE p."+postSet.getTableId()+"=tm."+postSet.getTableId();
	String del = "DELETE FROM "+matchTable.getTableName();

	String qCpy = "INSERT INTO "+matchTable.getTableName()+"_COPY ("+postSet.getTableId()+","+refSet.getTableId()+")";
	qCpy += " SELECT tm."+postSet.getTableId()+",tm."+refSet.getTableId()+" FROM "+trueMatchesTableName+" tm, "+postSet.getTableName()+" p WHERE p."+postSet.getTableId()+"=tm."+postSet.getTableId();
	String delCpy = "DELETE FROM "+matchTable.getTableName()+"_COPY";
	try {
	    postSet.openConnection();
	    (postSet.getStatement()).executeUpdate(del);
	    (postSet.getStatement()).executeUpdate(q);
	System.err.println("Q IS: "+q);
	    (postSet.getStatement()).executeUpdate(delCpy);
	    (postSet.getStatement()).executeUpdate(qCpy);
	    postSet.closeConnection();
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
    }
    
}
