package edu.isi.integration.Phoebus;

import java.util.*;
import java.io.*;
import java.sql.*;
import com.wcohen.secondstring.*;
import com.wcohen.secondstring.expt.*;
import edu.isi.integration.Commons.DataSet;

public class Phoebus {
    private String classificationFile;
    private String segmentationFile;
    private String mode;
    private HashMap clusters; //this is the cluster of candidates for a post
    private String candidateFile;
    private String directoryPath;//path to things like the exe and the files
    private boolean isOneToOne;//set this to true if you want your matches to enforce 1-1
    //otherwise they are 1-many
    
    public void setOneToOne(boolean b) {
	isOneToOne = b;
    }
    public void setClassFile(String cf) { classificationFile = cf; }
    public void setSegFile(String sf) { segmentationFile = sf; }
    public void setMode(String m) { mode = m; }
    public void setPath(String s) { directoryPath = s; }

    public String getClassFile() { return classificationFile; }
    public String getSegFile() { return segmentationFile; }
    public String getMode() { return mode; }

    public void setCandidateFile(String filename) {
	candidateFile = filename;
    }

    public void clearCandidates() {
	clusters.clear();
    }

    public void setCandidates(HashMap cands) {
	clusters = cands;
    }

    public void buildSVMClassData(DataSet pt, DataSet rst, DataSet matchTable) {
	//System.out.println("writing classification file...");
	Vector refAttrs =  rst.getAttributes();

	//first, we set up the transformations so that they
	//can accumulate their statistics
	int numTrans = refAttrs.size()+1;
	Vector stringWrapsForTrans[] = new Vector[numTrans]; //SecondString requires string wrappers for doing its calcs,
	//so we make vectors of the string wrappers	
	Transformations trans[] = new Transformations[numTrans]; //these hold the different functions that we use

	//init these arrays
	for(int x = 0; x < trans.length; x++) {
	    stringWrapsForTrans[x] = new Vector();
	    trans[x] = new Transformations();
	}

	//now we add each post to the string wraps objects
	//these are used for accumulating stats and stuff, so we need to add the post to each one, since we have each of them for each attribute of the reference set
	pt.setTableVals();
	HashMap posts = pt.getTableVals();
	Iterator iterator = posts.keySet().iterator();
	while (iterator.hasNext()) {
	    String key = (String)iterator.next();
	    HashMap currPost = (HashMap)posts.get(key);	   
	    String postVal = (String)currPost.get("post");
	    if (postVal != null) {
		for(int ix = 0; ix < stringWrapsForTrans.length; ix++) {
		    stringWrapsForTrans[ix].add(new StringWrapper(postVal));
		}
		//System.out.println("ADDED: "+key+" -- "+pVal);
	    }
	}
	
	rst.setTableVals();
	HashMap refSet = rst.getTableVals();
	Vector refAttrSchema = rst.getAttributes();
	Iterator i2 = refSet.keySet().iterator();
	while(i2.hasNext()) {
	    String k = (String)i2.next();
	    HashMap rattrs = (HashMap)refSet.get(k);
	    String rsAttrs = "";//string used to hold the refset attributes
	    for(int z = 0; z < refAttrSchema.size(); z++) {
		String ba = (String)refAttrSchema.get(z);
		rsAttrs += (String)rattrs.get((String)ba)+" ";
	    }
	    rsAttrs = rsAttrs.trim();

	    //add all of the attributes to the string wraps arrays for accumulating stats
	    for(int ii = 0; ii < stringWrapsForTrans.length; ii++) {
		if(ii == (stringWrapsForTrans.length - 1)) {
		    stringWrapsForTrans[ii].add(new StringWrapper(rsAttrs));
		    //last one uses full string
		}
		else {		    
		    //rest use single attr
		    String aName = (String)refAttrSchema.get(ii);
		    String aVal = (String)rattrs.get((String)aName);
		    stringWrapsForTrans[ii].add(new StringWrapper(aVal));
		} 
	    }
	}

	//accumulate all of the statistics for the distance functions
	for(int w = 0; w < trans.length; w++) {
	    trans[w].accumulateStats(stringWrapsForTrans[w]);
	}

	//now, get all of the blocking pairs, which are calculated by BSL and written to a file...
	HashMap clusters = loadCandidatesFromFile(candidateFile);

	//now we have our clusters, let's write our file for the SVM to classify!
	try {
	    BufferedWriter outFold1 = new BufferedWriter(new FileWriter(directoryPath+classificationFile));
	    Iterator clusterIter = clusters.keySet().iterator();
	    matchTable.openConnection(); //open it now, so we can speed up the checks...
	    while(clusterIter.hasNext()) {
		String currPostId =  (String)clusterIter.next();
		Vector refTuples = (Vector)clusters.get((String)currPostId);
		
		/***
		    LOAD UP A RECORD OBJECT FOR THE POST...for this, we have the attribute post...
		    LOAD UP A VECTOR OF RECORD OBJECTS FOR THE REF SET
		***/
		HashMap currPost = (HashMap)posts.get((String)currPostId);
		if(currPost == null) {
		    System.out.println("NULL FOR CURR POST ID: "+currPostId);
		}

		//will hold the output, and allVals will hold the cocatenation of them
		String vals[] = new String[refTuples.size()];
		String allVals[] = new String[refTuples.size()];
		
		//need initialization
		for(int iz = 0; iz < refTuples.size(); iz++) {
		    vals[iz] = "";
		    allVals[iz] = "";
		}
		
		//then write if it's a match or not, if you are training
		for(int zz = 0; zz < refTuples.size(); zz++) {
		    String refIdVal = (String)refTuples.get(zz);
		    if(mode.equals("train") || mode.equals("train_large")) {
			if(isMatch(pt, currPostId,rst,refIdVal,matchTable)) {
			    allVals[zz] = "+1 ";
			}
			else {
			    allVals[zz] = "-1 ";
			}
		    }
		    else {
			//means we are in testing mode
		  	allVals[zz] = "0 ";
		    }
		}

		//and now get the scores for the tuples		   
		int end = 0;//these are indexes for making file
		int start = 1;		

		//we iterate through the schema of the reference set, and test the post against each attribute
		String matchStr = (String)currPost.get("post");
		Vector refConcats = new Vector(); //we can build up the concatenations as we go. to use later...
		for(int r = 0; r < refAttrSchema.size(); r++) {
		    Vector currRefStrings = new Vector();
		    for(int xx = 0; xx < refTuples.size(); xx++) {
			HashMap currRefAttrs = (HashMap)refSet.get((String)refTuples.get(xx));	
			String refStr = (String)currRefAttrs.get((String)refAttrSchema.get(r));
			currRefStrings.add((String)refStr);

			//add it to the concatenation...
			if(r == 0) {
			    refConcats.add((String)refStr);
			}
			else {
			    String crs = (String)refConcats.get(xx);
			    crs += " "+refStr;
			    refConcats.set(xx, (String)crs); 
			}
		    }
		    if(r == 0) {
			//vals = trans[r].calculateValuesRetStrings(currRefStrings, matchStr);
			vals = trans[r].calculateMaxValuesRetStrings(currRefStrings, matchStr);
			for(int c = 0; c < vals.length; c++) {
			    String brok[] = vals[c].split("\\|");
			    end = brok.length;
			    for(int d = 0; d < brok.length; d++) {
				allVals[c] += start+":"+brok[d]+" ";
				start++;
			    }
			    start = 1;
			}
		    }
		    else {
			//vals = trans[r].calculateValuesRetStrings(currRefStrings, matchStr);
			vals = trans[r].calculateMaxValuesRetStrings(currRefStrings, matchStr);
			start = end+1;
			int tmp = start;
			//now add the values to the allVals vector
			for(int c = 0; c < vals.length; c++) {
			    String brok[] = vals[c].split("\\|");
			    end = brok.length+start-1;
			    for(int d = 0; d < brok.length; d++) {
				allVals[c] += start+":"+brok[d]+" "; 
				start++;
			    }
			    start = tmp;
			}
		    }    
		}

		//now we do it for the full concatenation of attributes...
		//vals = trans[numTrans-1].calculateValuesRetStrings(refConcats, matchStr);
		vals = trans[numTrans-1].calculateMaxValuesRetStrings(refConcats, matchStr);
		start = end+1;
		int tmp = start;
		//now add the values to the allVals vector
		for(int c = 0; c < vals.length; c++) {
		    String brok[] = vals[c].split("\\|");
		    end = brok.length+start-1;
		    for(int d = 0; d < brok.length; d++) {
			allVals[c] += start+":"+brok[d]+" "; 
			start++;
		    }
		    start = tmp;
		}

		//Now we deduplicate some of the feature vectors
		//which we can do b/c SV doesn't care about repeats...
		if(mode.equals("train_large")) {
		    Vector v = new Vector();
		    for(int a = 0; a < allVals.length; a++) {
			if(!v.contains((String)allVals[a])) {
			    v.add((String)allVals[a]);
			}
			else {
			    allVals[a] = "REPEAT";
			}
		    }
		    v.clear();
		}

		//now, write out to the file and add comments of the 
		for(int w = 0; w < allVals.length; w++) {
		    if(!allVals[w].equals("REPEAT")) {
			outFold1.write(allVals[w]+"#_#"+currPostId+"|"+(String)refTuples.get(w));
			outFold1.newLine();
		    }
		}

	    } //finished this cluster
	    outFold1.close();
	    //System.out.println("WROTE CLASSIFICATION "+mode+" FILE...");
	    matchTable.closeConnection(); //close our match table
	} //finished all of the file
	catch(Exception e) { e.printStackTrace();}

    } //END OF CLASSIFICATION FILE CREATION

    public HashMap getBlockClusters(Vector blocks) {
	HashMap clusters = new HashMap();
	for(int i = 0; i < blocks.size(); i++) {
	    String pair = ((String)blocks.get(i));
	    String brokPair[] = pair.split("\\|");
	    String matchID = brokPair[1];
	    String refID = brokPair[0];
	    
	    //first, check to see if there is already a vector there, if not, then make one and add value
	    //other wise, just append value
	    if(clusters.containsKey((String)matchID)) {
		Vector refs = (Vector)clusters.get((String)matchID);
		clusters.remove((String)matchID);
		refs.add((String)refID);
		clusters.put((String)matchID, (Vector)refs);
	    }
	    else {
		Vector refs = new Vector();
		refs.add((String)refID);
		clusters.put((String)matchID, (Vector)refs);
	    }
	}
	return clusters;
    }

    public void trainSVMClassifier(String trainingData, String modelName) {
	//System.out.println("Training classifier...");
	String ls_str;
	try { 
	    //String cmd = "\""+directoryPath+"svm_learn.exe\" \"-t\" \"2\" \""+directoryPath+trainingData+"\" \""+directoryPath+modelName+"\"";
	    String cmd = "\""+directoryPath+"svm_learn.exe\" \""+directoryPath+trainingData+"\" \""+directoryPath+modelName+"\"";
	    //System.out.println("CMD: "+cmd);
	    Process p = Runtime.getRuntime().exec(cmd);
	    DataInputStream ls_in = new DataInputStream(p.getInputStream());
	    try {
		while ((ls_str = ls_in.readLine()) != null) {
		    //System.out.println(ls_str);
		}
	    } 
	    catch (Exception e) {
		e.printStackTrace();
	    }
	    p.waitFor();
	    p.destroy();
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
	//System.out.println("Done training...");
    }

    public void trainSVMSegmenter(String trainingData, String modelName) {
	//System.out.println("Training segmenter...");
	String ls_str;
	try {
	    String cmd = "\""+directoryPath+"svm_multiclass_learn.exe\" \""+directoryPath+trainingData+"\" \""+directoryPath+modelName+"\"";
	    Process p = Runtime.getRuntime().exec(cmd);
	    DataInputStream ls_in = new DataInputStream(p.getInputStream());
	    try {
		while ((ls_str = ls_in.readLine()) != null) {
		    //System.out.println(ls_str);
		}
	    } 
	    catch (Exception e) {
		e.printStackTrace();
	    }
	    p.waitFor();		
	    p.destroy();

	}
	catch(Exception e) {
	    e.printStackTrace();
	}
	//System.out.println("Done training segmenter...");
    }

    public void runSVMClassifier(String testData, String modelName, String predictionFile) {
	//System.out.println("Running classifier...");
	String ls_str;
	try {
	    String cmd = "\""+directoryPath+"svm_classify.exe\" \""+directoryPath+testData+"\" \""+directoryPath+modelName+"\" \""+directoryPath+predictionFile+"\"";
	    Process p = Runtime.getRuntime().exec(cmd);
	    DataInputStream ls_in = new DataInputStream(p.getInputStream());
	    try {
		while ((ls_str = ls_in.readLine()) != null) {
		    //System.out.println(ls_str);
		}
	    } 
	    catch (Exception e) {
		e.printStackTrace();
	    }
	    p.waitFor();		
	    p.destroy();
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
	//System.out.println("Done classifying...");
    }

    public void runSVMSegmenter(String testData, String modelName, String predictionFile) {
	//System.out.println("Running segmenter...");
	String ls_str;
	try {
	    String cmd = "\""+directoryPath+"svm_multiclass_classify.exe\" \""+directoryPath+testData+"\" \""+directoryPath+modelName+"\" \""+directoryPath+predictionFile+"\"";
	    Process p = Runtime.getRuntime().exec(cmd);
	    DataInputStream ls_in = new DataInputStream(p.getInputStream());
	    try {
		while ((ls_str = ls_in.readLine()) != null) {
		    //System.out.println(ls_str);
		}
	    } catch (Exception e) {
		e.printStackTrace();
	    }
	    p.waitFor();		
	    p.destroy();
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
	//System.out.println("Done segmenting...");
    }

    public void getMatches(String classFile, String predsFile, DataSet rst, DataSet pt, DataSet matchTable) {
	//class file holds the vector and which id looking for
	//the preds file says if its a match or not
	int lineCount = 0;
	try {
	    //declared here rather than when do insert, to speed things up!
	    matchTable.openConnection();
	    BufferedReader inPred = new BufferedReader(new FileReader(directoryPath+predsFile));
	    BufferedReader inClass = new BufferedReader(new FileReader(directoryPath+classFile));
	    String strP = "";
	    String strC = "";
	    String oldID = "-1";
	    double maxVal = -999.999;
	    String maxRef = "-1";
	    Vector maxRefs = new Vector();
	    int maxCount = 0; //only one ref match allowed

	    while(((strP = inPred.readLine()) != null)) {
		//&& ((strC = inClass.readLine()) != null)) {
		strC = inClass.readLine();
		lineCount++;
		String[] broken = strC.split("#_#");
		String[] vals = broken[1].split("\\|");
		String id = vals[0];
		String ref = vals[1];

		//NEW STUFF
		if(oldID.equals("-1")) {
		    oldID = id;
		    maxVal = (new Double(strP).doubleValue());
		    maxCount = 1;
		    maxRefs.add((String)ref);
		}
		else if(!id.equals(oldID) && (!oldID.equals("-1")) ) {
		    if(isOneToOne) {
			if(maxCount == 1) {
			    for(int i = 0; i < maxRefs.size(); i++) {
				String mRef = (String)maxRefs.get(i);
				insertMatch(matchTable, oldID, mRef, rst, pt);
			    }
			}
		    }
		    else {
			//System.out.println("FOR :"+oldID+"WE HAVE: "+maxRefs);
			for(int ii = 0; ii < maxRefs.size(); ii++) {
			    String mRef = (String)maxRefs.get(ii);
			    insertMatch(matchTable, oldID, mRef, rst, pt);
			}
		    }
		    maxRefs.clear();
		    oldID = id;
		    maxRef = ref;
		    maxRefs.add((String)ref);
		    maxVal = (new Double(strP).doubleValue());
		    //System.out.println("ID UPDATE TO..."+oldID+" WITH SCORE: "+maxVal);
		    maxCount = 1;
		}
		else {		 
		    double tmp = (new Double(strP).doubleValue());
		    if(tmp == maxVal) { 
			//System.out.println("FOUND SAME SCORE"+maxVal+" FOR ID: "+oldID);
			maxCount++; 
			maxRefs.add((String)ref);
		    }
		    if(tmp > maxVal) {
			maxRefs.clear();
			maxRefs.add((String)ref);
			maxVal = tmp;
			maxRef = ref;
			maxCount=1;
			//System.out.println("SCORE UPDATE"+maxVal+" FOR ID: "+oldID);
		    }
		}
	    }

	    if(isOneToOne) {
		if(maxCount == 1) {
		    for(int j = 0; j < maxRefs.size(); j++) {
			String mRef = (String)maxRefs.get(j);
			insertMatch(matchTable, oldID, mRef, rst, pt);
		    }
		}
	    }
	    else {
		for(int jj = 0; jj < maxRefs.size(); jj++) {
		    String mRef = (String)maxRefs.get(jj);
		    insertMatch(matchTable, oldID, mRef, rst, pt);
		}
	    }

	    inPred.close();
	    inClass.close();
	    matchTable.closeConnection();
	}
	catch (Exception e) {
	    System.out.println("ERROR ON LINE "+lineCount);
	    e.printStackTrace();
	}
    }

    public void insertMatch(DataSet matchTable, String idS, String refS,
			    DataSet rst, DataSet pt) {
	//first we need to get the table information
	if(refS.equals("-1")) {
	    System.out.println("UH OH...");
	    return;
	    //the above can happen when the first post has no match, so the refid is never updated and stays at the
	    //initial placeholder of -1
	}
	//System.out.println("MATCH OF ID: "+idS+" AND REF: "+refS);

	//String query = "UPDATE "+matchTable.getTableName()+" SET "+rst.getTableId()+"='"+refS+"' WHERE "+pt.getTableId()+"='"+idS+"' AND "+rst.getTableId()+"='0'";
	String query = "INSERT INTO "+matchTable.getTableName()+" ("+pt.getTableId()+","+rst.getTableId()+") VALUES ('"+idS+"','"+refS+"')";
	try {
	    (matchTable.getStatement()).executeUpdate(query);
	}
	catch (Exception e) {
	    e.printStackTrace();
	    System.out.println("FAILED ON: "+query);
	}
    }

    public void buildSVMSegmentationData(DataSet pt, DataSet rst, DataSet matchTable, HashMap commonScores) {
	//System.out.println("Writing segmentation file...");
	//first, we only need to process the posts from the post table that have a match in the reference set
	//now we get the posts and the reference set information (and common scores too...)
	HashMap postsWithMatches = getPostsWithMatches(pt, rst, matchTable);

	Vector postAttrNames = pt.getAttributes();
	/***
	    HERE WE NEED ONLY COMMON SCORES AND REF SET SCHEMA, NOT POST
	***/
	postAttrNames.remove((String)"post");

	HashMap refSet = rst.getTableVals();
	if(refSet == null) {
	    rst.setTableVals();
	    refSet = rst.getTableVals();
	}

	//Vector refSetAttrs = rst.getAttributes();
	try {
	    BufferedWriter segFile = new BufferedWriter(new FileWriter(directoryPath+segmentationFile));
	    matchTable.openConnection(); //open the match table connection so we can reuse it later, to speed things up
	    //now, we iterate through the posts and compare each token of a post to the matching reference set member's 
	    //attributes, and the common scores
	    Iterator iterator = postsWithMatches.keySet().iterator();
	    while (iterator.hasNext()) {
		String currPostId = (String)iterator.next(); //iterate through posts
		HashMap currPost = (HashMap)postsWithMatches.get((String)currPostId); //get the current post for our id
		String post = (String)currPost.get("post"); 

		//for labeling training data and for getting matchign ref set member's id
		String refId = getRefId(currPostId, pt, rst, matchTable);
		HashMap refSetAttrs = (HashMap)refSet.get((String)refId);
		if(refSetAttrs == null) {
		    System.out.println("NULL FOR ID: "+currPostId+" WITH REF: "+refId);
		    return;
		}

		//now we will tokenize the string and do the comparison to each attr and common scores
		//first we remove extra spaces from the post and then we tokenize
		post = post.replaceAll("\\s+"," ");
		String brok[] = post.split(" ");
		Transformations trans = new Transformations();
		trans.setStatsFlag();
		//String outputLine = "";
		//String comment = " #_#"+post;

		for(int z = 0; z < brok.length; z++) {
		    String outputLine = "";
		    int counter = 1;
		    //first, write down the true label
		    if(mode.equals("train") || mode.equals("train_large")) {
			outputLine = getTrueLabel(brok[z],postAttrNames,currPost)+" ";
		    }
		    else {
			outputLine = (postAttrNames.size()+1)+" ";
		    }

		    //now do the common score stuff
		    Iterator commScoreIterator = commonScores.keySet().iterator();
		    while(commScoreIterator.hasNext()) {
			String csAttr = (String)commScoreIterator.next();
			String commScore = (String)commonScores.get((String)csAttr);
			if(brok[z].matches(commScore)) {
			    outputLine += counter+":10 ";
			}
			else {
			    outputLine += counter+":0 ";
			}
			counter++;
		    }

		    //now do comparisons to all attributes of the reference set
		    Iterator refSetAttrIter = refSetAttrs.keySet().iterator();
		    while(refSetAttrIter.hasNext()) {
			String rsAttrName = (String)refSetAttrIter.next();
			String rsAttrVal = (String)refSetAttrs.get((String)rsAttrName);
			outputLine += counter+":"+trans.getSmithWatermanScore(rsAttrVal,brok[z])+" ";
			counter++;
			outputLine += counter+":"+trans.getJaroWinklerScore(rsAttrVal,brok[z])+" ";
			counter++;
			outputLine += counter+":"+trans.getLevensteinScore(rsAttrVal,brok[z])+" ";
			counter++;
			outputLine += counter+":"+trans.getSoundexScore(rsAttrVal,brok[z])+" ";
			counter++;
			outputLine += counter+":"+trans.getStemScore(rsAttrVal,brok[z])+" ";
			counter++;
		    }
		    outputLine += " #_#"+currPostId+"|"+brok[z];
		    segFile.write(outputLine);
		    segFile.newLine();
		    segFile.flush();
		}//end loop over each token for a post
	    }//end iteration through posts
	    segFile.close();
	    matchTable.closeConnection();
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
	//System.out.println("WROTE SEGMENTATION "+mode+" FILE...");
    }

    public String getTrueLabel(String token, Vector attributeNames, HashMap attributeVals) {
	for(int i = 0; i < attributeNames.size(); i++) {
	    String attrName = (String)attributeNames.get(i);
	    String refAttr = (String)attributeVals.get((String)attrName);
	    if(refAttr == null) { refAttr = ""; }
	    if(refAttr.trim().length() > 0 ) {
		String refToks[] = refAttr.split("\\s");
		for(int j = 0; j < refToks.length; j++) {
		    if(refToks[j].equals(token.trim())) { 
			return (i+1)+""; 
		    }
		}
	    }
	}
	return (attributeNames.size()+1)+"";
    }

    public void getLabeledTokens(String segFile, String predsFile, DataSet pt,  HashMap commonScores) {
	int lineCount = 0;
	//System.out.println("Labeling extracted tokens...");
	try {
	    //put the db connection stuff here, since make lots of db calls, will speed it up!
	    Class.forName("sun.jdbc.odbc.JdbcOdbcDriver");
	    Connection con = DriverManager.getConnection(pt.getURL(),pt.getLogin(),pt.getPass());
	    Statement stmt = con.createStatement();

	    BufferedReader inPred = new BufferedReader(new FileReader(directoryPath+predsFile));
	    BufferedReader inSeg = new BufferedReader(new FileReader(directoryPath+segFile));
	    String strP = "";
	    String strS = "";
	    //int junkLabel = (pt.getAttributes()).size()+(commonScores.keySet()).size();
	    int junkLabel = (pt.getAttributes()).size() + 1;
	    while((strP = inPred.readLine()) != null) {
		lineCount++;
		strS = inSeg.readLine();
		if(strS != null) {
		    ////System.out.println(strS);
		    String[] broken = strS.split("#_#");
		    String[] vals = broken[1].split("\\|");
		    if(vals.length > 1) {
			String postId = vals[0];
			String token = vals[1];
			int label = (new Integer(strP).intValue());
			if(label != junkLabel) {
			    updateField(postId,token,label-1,pt,stmt,commonScores);
			}
		    }
		}
	    }
	    inPred.close();
	    inSeg.close();
	    stmt.close();
	    //con.commit();
	    con.close();
	}
	catch(Exception e) {
	    System.out.println("ON LINE: "+lineCount);
	    e.printStackTrace();
	}
	//System.out.println("Finished labeling extracted tokens...");
    }

    public void updateField(String postId, String token, int label, DataSet pt, Statement stmt,HashMap commonScores) {
	Vector pAttrNames = pt.getAttributes();
	Set commonScoreAttrs = commonScores.keySet();
	pAttrNames.addAll((Collection)commonScoreAttrs);
	//System.out.println("pAttrNames: "+pAttrNames);
	if(label > pAttrNames.size()-1) {
	    //means it was labeled junk, so we do nothing with it
	    return;
	}
	String fieldName = (String)pAttrNames.get(label);
	//System.out.println("LABEL: "+label+"  FIELD: "+fieldName);
	String tableName = pt.getTableName();
	String tableId = pt.getTableId();

	try {
	    String qq = "SELECT "+fieldName+" FROM "+tableName+" WHERE "+tableId+"='"+postId+"'";
	    ResultSet res = stmt.executeQuery(qq);
	    String found = "";
	    while(res.next()) {
		found = res.getString(fieldName);
	        if(found.equals("0")) {
		   found = "";
		}
	    }
	    token = token.replaceAll("'","\\\\'");
	    found += " "+token;
	    found = found.trim();
	    String query = "UPDATE "+tableName+" SET "+fieldName+"='"+found+"' WHERE "+tableId+"='"+postId+"'";
	    //System.out.println(query);
	    stmt.executeUpdate(query);
	    //stmt.close();
	    //con.close();
	}
	catch (Exception e) {
	    e.printStackTrace();
	}
    }

    private String cleanField(String ref, String post) {
	if(ref == null) {
	    //somehow we don't have a ref for it!
	    return post;
	}

	if(ref.trim().length() == 0) {
	    //some weird empty string
	    return post;
	}

	if(post.trim().length() == 0) {
	    //some weird all whitespace post!
	    return post.trim();
	}

	Jaccard js = new Jaccard();
	JaroWinkler lv = new JaroWinkler();

	String brokenpost[] = post.split(" ");

	Vector removedToks = new Vector();
	int removeIndex = -1;
	String savedToks = "";
	String temp = "";

	double baselineJW = lv.score(lv.prepare(ref.toLowerCase()),lv.prepare(post.toLowerCase()));
	double baselineJS = js.score(js.prepare(ref.toLowerCase()),js.prepare(post.toLowerCase()));
	double maxJW = 0.0;
	double maxJS = 0.0;

	for(int i = 0; i < brokenpost.length; i++) {
	    for(int j = 0; j < brokenpost.length; j++) {
		if(i != j) {
		    temp += brokenpost[j]+" ";
		}
	    }
	    temp = temp.trim();  
	    double scoreJW = lv.score(lv.prepare(ref.toLowerCase()),lv.prepare(temp.toLowerCase()));
	    double scoreJS = js.score(js.prepare(ref.toLowerCase()),js.prepare(temp.toLowerCase()));

	    if( (scoreJW > baselineJW) && (scoreJS > baselineJS) ) {
	    	if( (scoreJW > maxJW) && (scoreJS > maxJS)) {
		    maxJW = scoreJW;
		    maxJS = scoreJS;
		    removeIndex = i;
		}
	    }
	    
	    temp = "";
	}

	for(int z = 0; z < brokenpost.length;z++) {
	    if(z != removeIndex) {
		savedToks += brokenpost[z]+" ";
	    }
	}
	savedToks = savedToks.trim();

	if(removeIndex != -1) {
	    cleanField(ref, savedToks);
	}
	return savedToks;
    }
    
    public void postProcessFields(DataSet pt, DataSet rst, DataSet matchTable) {
	//first, get all of the posts that had a match, the attribute names and the reference set
	//pt.reset();
	HashMap postsWithMatches = getPostsWithMatches(pt, rst, matchTable);
	//rst.setTableVals();
	//now, we iterate through the posts and compare the analogue ref set attributes
	//that means, first we check each attribute and see if it's in this ref set...
	Iterator iterator = postsWithMatches.keySet().iterator();
	try {
	    Class.forName("sun.jdbc.odbc.JdbcOdbcDriver");
	    Connection cond = DriverManager.getConnection(pt.getURL(),pt.getLogin(),pt.getPass());
	    Statement stmt = cond.createStatement();
	    matchTable.openConnection();
	    pt.openConnection();
	    while (iterator.hasNext()) {
		String currPostId = (String)iterator.next(); //iterate through posts
		HashMap currPost = (HashMap)postsWithMatches.get((String)currPostId); //get the current post for our id
		
		//now we need the Ref set member for our post
		String refId = getRefId(currPostId,pt,rst,matchTable);
		HashMap refSetMember = (HashMap)((rst.getTableVals()).get((String)refId));

		//iterate through these and see if in the reference set
		Iterator pAttrsIter = currPost.keySet().iterator();
		while(pAttrsIter.hasNext()) {
		    String pAttrName = (String)pAttrsIter.next();		
		    //a little sanity check to make sure the reference set has this attribute...
		    if(refSetMember != null) {
			if(refSetMember.containsKey((String)pAttrName)) {
			    String refAttrValue = (String)refSetMember.get((String)pAttrName);
			    //String postAttrValue = ((String)currPost.get((String)pAttrName)).trim();
			    String postAttrValue = getPostAttrValue(pAttrName,currPostId,pt,pt.getStatement());
			    String cleanedField = cleanField(refAttrValue, postAttrValue);
			    if(!postAttrValue.equals(cleanedField)) {
				//System.out.println("CLEANED: "+postAttrValue+" TO: "+cleanedField);
				updateCleanedField(pt,currPostId,pAttrName,cleanedField,stmt);
			    }
			}
		    }
		}
	    }
	    pt.closeConnection();
	    matchTable.closeConnection();
	    //cond.commit();
	    stmt.close();
	    cond.close();
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
    }

    private void updateCleanedField(DataSet pt, String postId, String attrName, String cleanedValue, Statement stmt) {
	String query = "UPDATE "+pt.getTableName()+" SET "+attrName+"='"+cleanedValue+"' WHERE "+pt.getTableId()+"='"+postId+"'";
	try {
	    stmt.executeUpdate(query);
	}
	catch (Exception e) {
	    e.printStackTrace();
	}
    }

    public HashMap loadCandidatesFromFile(String filename) {
	HashMap cands = new HashMap();
	int line = 0;
	try {
	    BufferedReader in = new BufferedReader(new FileReader(directoryPath+filename));
	    String str = "";
	    while(((str = in.readLine()) != null)) {
		line++;
		String toks[] = str.split("\\|");
		String key = toks[0];
		String val = toks[1];
		if(cands.containsKey((String)key)) {
		    Vector v = (Vector)cands.get((String)key);
		    v.add((String)val);
		    cands.remove((String)key);
		    cands.put((String)key, (Vector)v);
		}
		else {
		    Vector v = new Vector();
		    v.add((String)val);
		    cands.put((String)key, (Vector)v);
		}
	    }
	}
	catch(Exception e) {
	    System.out.println("ERROR ON LINE: "+line);
	    e.printStackTrace();
	}
	return cands;
    }

    public boolean isMatch(DataSet set1, String set1id, 
			   DataSet set2, String set2id, 
			   DataSet matchTable) {
	String q = "SELECT "+matchTable.getTableName()+"."+set1.getTableId()+","+matchTable.getTableName()+"."+set2.getTableId();
	q += " FROM "+matchTable.getTableName();
	q += " WHERE "+set1.getTableId()+"='"+set1id+"' AND "+set2.getTableId()+"='"+set2id+"'";
	int numMatches = 0;
	try {
	    //we don't do count distinct b/c access is grumpy about it
	    ResultSet rs = (matchTable.getStatement()).executeQuery(q);
	    while (rs.next()) {
		numMatches++;
	    }
	}
	catch(Exception e) {
	    e.printStackTrace();
	    System.out.println("FAILED ON: "+q);
	    //System.exit(0);
	}
	if(numMatches == 0) {
	    return false;
	}
	return true;
    }

    private HashMap getPostsWithMatches(DataSet postSet, DataSet refSet, DataSet matchTable) {
	//the records in the match table with the reference set id set to 0 are ones that weren't matched to anything
	//so, we just find those in the post set and delete them...
	//System.out.println(query);
	String query = "SELECT "+postSet.getTableId()+" FROM "+postSet.getTableName()+" WHERE "+postSet.getTableName()+"."+postSet.getTableId()+" NOT IN (SELECT DISTINCT "+postSet.getTableId()+" FROM "+matchTable.getTableName()+" WHERE "+refSet.getTableId()+" NOT LIKE '0')";
	try {
	    matchTable.openConnection();
	    ResultSet rs = (matchTable.getStatement()).executeQuery(query);
	    while (rs.next()) {
		//as it turns out, we already have a nice method to do this
		postSet.removeRecord(rs.getString(postSet.getTableId()));
	    }
	    matchTable.closeConnection();
	}
	catch(Exception e) {
	    e.printStackTrace();
	    System.out.println("FAILED ON: "+query);
	}
	return postSet.getTableVals();
    }

    private String getRefId(String currPostId, DataSet postSet, DataSet refSet, DataSet matchTable) {
	String refId = "";
	String query = "SELECT DISTINCT "+refSet.getTableId()+" FROM "+matchTable.getTableName();
	query += " WHERE "+postSet.getTableId()+"='"+currPostId+"' AND "+refSet.getTableId()+" NOT LIKE '0'";
	try {
	    ResultSet rs = (matchTable.getStatement()).executeQuery(query);
	    while (rs.next()) {
		refId = rs.getString(refSet.getTableId());
	    }
	}
	catch(Exception e) {
	    e.printStackTrace();
	    System.out.println("FAILED ON: "+query);
	}
	return refId;
    }

    public void removeTemporaryFile(String filename) {
	try {
	    boolean success = (new File(directoryPath+filename)).delete();
	    if(!success) {
		System.err.println("COULD NOT DELETE: "+filename);
	    }
	}
	catch(Exception e) {
	    System.err.println("FAILED DELETING: "+filename);
	    e.printStackTrace();
	}
    }

    public String getPostAttrValue(String attrName, String id, DataSet postSet, Statement stmt) {
	String ret = "";
	try {
	    String q = "SELECT "+attrName+" FROM "+postSet.getTableName()+" WHERE "+postSet.getTableId()+"='"+id+"'";
	    ResultSet rs = stmt.executeQuery(q);
	    while(rs.next()) {
		ret = rs.getString(attrName);
	    }
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
	return ret;
    }
}