package edu.isi.integration.Phoebus;

import org.apache.commons.codec.language.*;
import com.wcohen.secondstring.*;
import java.util.*;

public class Transformations {
    private SmithWaterman smith;
    private JaroWinkler jaro ;
    private DirichletJS jen ;
    private Jaccard j ;
    private CharJaccard cj ;
    private Levenstein l ;
    private JelinekMercerJS jen2 ;
    private int statsFlag; //to see if we've accumulated stats or not, if no, then distance functs won't work

    public Transformations() {
	smith = new SmithWaterman();
	jaro = new JaroWinkler();
	jen = new DirichletJS();
	j = new Jaccard();
	cj = new CharJaccard();
	l = new Levenstein();
	jen2 = new JelinekMercerJS();
	statsFlag = 0;
    }

    public int getStemScore(String s, String t) {
	//counts the number of matching stemms
	PorterStemmer p = new PorterStemmer();
	String brokS[] = s.split("\\s");
	String brokT[] = t.split("\\s");
	
	int stemCount = 0;
	try {
	    for(int i = 0; i < brokS.length; i++) {
		for(int j = 0; j < brokT.length; j++) {
		    String tmpS = brokS[i];
		    String tmpT = brokT[j];
		    String stemmedS = p.stem(tmpS);
		    if(tmpT.equalsIgnoreCase(stemmedS)) {
			stemCount++;
		    }
		    else {
			String stemmedT = p.stem(tmpT);
			if(stemmedT.equalsIgnoreCase(stemmedS)) {
			    stemCount++;
			}
		    }
		    
		}
	    }
	}
	catch(Exception e) {}
	return stemCount;
    }

    public int getSoundexScore(String s, String t) {
	//gets the total soundex score for all tokens
	int soundexScore = 0;
	try {
	    String brokS[] = s.split("\\s");
	    String brokT[] = t.split("\\s");
	    Soundex sdx = new Soundex();
	    for(int i = 0; i < brokS.length; i++) {
		for(int j = 0; j < brokT.length; j++) {
		    String tmpS = brokS[i];
		    String tmpT = brokT[j];
		    //remove all non-alphas
		    tmpS = tmpS.replaceAll("\\W","");
		    tmpT = tmpT.replaceAll("\\W","");
		    soundexScore += sdx.difference(tmpS, tmpT); 
		}
	    }
	}
	catch(Exception e) { 
	    e.printStackTrace();
	}
	return soundexScore;	
    }

    public void setStatsFlag() { statsFlag = 1; } //used when you don't need to actually do it, but need to set to call functs

    public void accumulateStats(Vector fullSet) { //v should be a vector of stringwrappers
	jen.accumulateStatistics(fullSet.iterator());
	smith.accumulateStatistics(fullSet.iterator());
	jaro.accumulateStatistics(fullSet.iterator());
	j.accumulateStatistics(fullSet.iterator());
	l.accumulateStatistics(fullSet.iterator());
	jen2.accumulateStatistics(fullSet.iterator());
	statsFlag = 1;
    }

    public double getDirichletJenShanScore(String s, String t) {
	if(statsFlag == 1) {
	    return jen.score(jen.prepare(s),jen.prepare(t)); 
	}
	return 0.0;
    }

    public double getDirichletJenShanScore(StringWrapper ss, StringWrapper tt) {
	String s = ss.unwrap();
	String t = tt.unwrap();
	if(statsFlag == 1) {
	    return jen.score(jen.prepare(s),jen.prepare(t)); 
	}
	return 0.0;
    }

    public double getSmithWatermanScore(String s, String t) {
	if(statsFlag == 1) {
	    return smith.score(smith.prepare(s),smith.prepare(t));
	}
	return 0.0;
    }

    public double getSmithWatermanScore(StringWrapper ss, StringWrapper tt) {
	String s = ss.unwrap();
	String t = tt.unwrap();
	if(statsFlag == 1) {
	    return smith.score(smith.prepare(s),smith.prepare(t));
	}
	return 0.0;
    }

    public double getJaroWinkler2Score(String s, String t) {
	Level2JaroWinkler jwl2 = new Level2JaroWinkler();
	if(statsFlag == 1) {
	    double v = jwl2.score(s,t);
	    if((v+"").equals("-Infinity")) {
		v = -2.0;
	    }
	    return v;
	}
	return 0.0;
    }

    public double getMaxJaroWinklerScore(String ref, String tok) {
	double maxVal = 0.0;
	String splitVal[] = ref.split(" ");
	for(int i = 0; i < splitVal.length; i++) {
	    double v = jaro.score(splitVal[i],tok);
	    if(v > maxVal) {
		maxVal = v;
	    }
	}
	if((maxVal+"").equals("-Infinity")) {
	    maxVal = -2.0;
	}
	return maxVal;
    }

    public double getJaroWinklerScore(String s, String t) {
	if(statsFlag == 1) {
	    return jaro.score(jaro.prepare(s),jaro.prepare(t));
	}
	return 0.0;
    }

    public double getJaroWinklerScore(StringWrapper ss, StringWrapper tt) {
	String s = ss.unwrap();
	String t = tt.unwrap();
	if(statsFlag == 1) {
	    return jaro.score(jaro.prepare(s),jaro.prepare(t));
	}
	return 0.0;
    }

    public double getCharJaccardScore(String s, String t) {
	if(statsFlag == 1) {
	    return cj.score(cj.prepare(s),cj.prepare(t));
	}
	return 0.0;
    }

    public double getJaccScore(String s, String t) {
	double sc =j.score(j.prepare(s),j.prepare(t));
	if(!Double.isNaN(sc)) {
	    return sc;
	}
	return 0;
    }
    public double getTFIDFScore(String s, String t) {
	TFIDF tf = new TFIDF();
	return tf.score(s,t);
    }

    public double getJaccardScore(String s, String t) {
	if(statsFlag == 1) {
	    return j.score(j.prepare(s),j.prepare(t));
	}
	return 0.0;
    }

    public double getJaccardScore(StringWrapper ss, StringWrapper tt) {
	String s = ss.unwrap();
	String t = tt.unwrap();
	if(statsFlag == 1) {
	    return j.score(j.prepare(s),j.prepare(t));
	}
	return 0.0;
    }

    public double getLevensteinScore(String s, String t) {
	if(statsFlag == 1) {
	    return l.score(l.prepare(s),l.prepare(t));
	}
	return 0.0;
    }

    public double getLevensteinScore(StringWrapper ss, StringWrapper tt) {
	String s = ss.unwrap();
	String t = tt.unwrap();
	if(statsFlag == 1) {
	    return l.score(l.prepare(s),l.prepare(t));
	}
	return 0.0;
    }

    public double getJelinekMercerJSScore(String s, String t) {
	if(statsFlag == 1) {
	    return jen2.score(jen2.prepare(s),jen2.prepare(t));
	}
	return 0.0;
    }

    public double getJelinekMercerJSScore(StringWrapper ss, StringWrapper tt) {
	String s = ss.unwrap();
	String t = tt.unwrap();
	if(statsFlag == 1) {
	    return jen2.score(jen2.prepare(s),jen2.prepare(t));
	}
	return 0.0;
    }

    public double getEqualsScore(String s, String t) {
	if(statsFlag == 1) {
	    if(s.equals(t)) {
		return 1.0;
	    }
	}
	return 0.0;
    }

    public double getSubstringScore(String s, String t) {
	if(statsFlag == 1) {
	    if(s.indexOf(t) > -1) {
		return 1.0;
	    }
	}
	return 0.0;
    }

    public String[] calculateValuesRetStrings(Vector refStrings, String matchStr) {
	String vals[] = new String[refStrings.size()];
	Vector maxJShanInd = new Vector(); //holds index of max (possibly maxes) of jensen shannon dist with maximal values
	Vector maxJaccInd = new Vector(); 
	//Vector maxCJaccInd = new Vector(); 
	Vector maxJShan2Ind = new Vector(); 
	Vector maxSmithInd = new Vector(); 
	Vector maxJaroInd = new Vector(); 
	Vector maxLevInd = new Vector(); 
	Vector maxSoundexInd = new Vector();
	Vector maxStemInd = new Vector();

	int matchIndex = -1;

	//firstly, get all of the maximum indexes for capturing this feature function
	//such as which one has the maximally similar jensen shannon distance
	double maxDistJen = 0.0;
	double maxDistJacc = 0.0;
	//double maxDistCJacc = 0.0;
	double maxDistJen2 = 0.0;
	double maxDistS = 0.0;
	double maxDistJaro = 0.0;
	double maxDistL = -0.1;
	double maxSoundex = 0.0;
	double maxStem = 0.0;

	for(int i = 0; i < refStrings.size(); i++) {
	    vals[i] = "";

	    String refStr = (String)refStrings.get(i);
	    double dJen = jen.score(jen.prepare(refStr),jen.prepare(matchStr));
	    double dJacc = j.score(j.prepare(refStr),j.prepare(matchStr));
	    //double dCJacc = cj.score(cj.prepare(refStr),cj.prepare(matchStr));
	    double dJen2 = jen2.score(jen2.prepare(refStr),jen2.prepare(matchStr));
	    double dS = smith.score(smith.prepare(refStr),smith.prepare(matchStr));
	    double dJaro = jaro.score(jaro.prepare(refStr),jaro.prepare(matchStr));
	    double dL = l.score(l.prepare(refStr),l.prepare(matchStr));
	    double dStem = -1;
	    try { 
		dStem = getStemScore(refStr, matchStr);
	    }
	    catch (Exception e) {}
	    double dSound = getSoundexScore(refStr, matchStr);
	    vals[i] = dJen+"|"+dJacc+"|"+dJen2+"|"+dS+"|"+dJaro+"|"+dL+"|"+dStem;
	}
	return vals;
    }

    public String[] calculateMaxValuesRetStrings(Vector refStrings, String matchStr) {
	String vals[] = new String[refStrings.size()];
	Vector maxJShanInd = new Vector(); //holds index of max (possibly maxes) of jensen shannon dist with maximal values
	Vector maxJaccInd = new Vector(); 
	//Vector maxCJaccInd = new Vector(); 
	Vector maxJShan2Ind = new Vector(); 
	Vector maxSmithInd = new Vector(); 
	Vector maxJaroInd = new Vector(); 
	Vector maxLevInd = new Vector(); 
	Vector maxSoundexInd = new Vector();
	Vector maxStemInd = new Vector();

	int matchIndex = -1;

	//firstly, get all of the maximum indexes for capturing this feature function
	//such as which one has the maximally similar jensen shannon distance
	double maxDistJen = 0.0;
	double maxDistJacc = 0.0;
	//double maxDistCJacc = 0.0;
	double maxDistJen2 = 0.0;
	double maxDistS = 0.0;
	double maxDistJaro = 0.0;
	double maxDistL = -0.1;
	double maxSoundex = 0.0;
	double maxStem = 0.0;

	for(int i = 0; i < refStrings.size(); i++) {
	    String refStr = (String)refStrings.get(i);
	    Integer iI = new Integer(i);	    
	    double dJen = jen.score(jen.prepare(refStr),jen.prepare(matchStr));
	    double dJacc = j.score(j.prepare(refStr),j.prepare(matchStr));
	    double dJen2 = jen2.score(jen2.prepare(refStr),jen2.prepare(matchStr));
	    double dS = smith.score(smith.prepare(refStr),smith.prepare(matchStr));
	    double dJaro = jaro.score(jaro.prepare(refStr),jaro.prepare(matchStr));
	    double dL = l.score(l.prepare(refStr),l.prepare(matchStr));
	    double dStem = -1;
	    try { 
		dStem = getStemScore(refStr, matchStr);
	    }
	    catch (Exception e) {}
	    double dSound = getSoundexScore(refStr, matchStr);

	    if( (dJen == maxDistJen) && (dJen > 0.0)) {
		maxJShanInd.add(iI);
	    }
	    if(dJen > maxDistJen) {
		maxDistJen = dJen;
		maxJShanInd.clear();
		maxJShanInd.add(iI);
	    }
	    if( (dS == maxDistS) && (dS > 0.0)) {
		maxSmithInd.add(iI);
	    }
	    if(dS > maxDistS) {
		maxDistS = dS;
		maxSmithInd.clear();
		maxSmithInd.add(iI);
	    }
	    if( (dStem ==  maxStem) && (dStem > 0.0)) {
		maxStemInd.add(iI);
	    }
	    if(dStem >  maxStem) {
		maxStem = dStem;
		maxStemInd.clear();
		maxStemInd.add(iI);
	    }
	      	      
	    if( (dSound ==  maxSoundex) && (dSound > 0.0)) {
		maxSoundexInd.add(iI);
	    }
	    if(dSound >  maxSoundex) {
		maxSoundex = dSound;
		maxSoundexInd.clear();
		maxSoundexInd.add(iI);
	    }
	    if( (dJacc == maxDistJacc) && (dJacc > 0.0)) {
		maxJaccInd.add(iI);
	    }
	    if(dJacc > maxDistJacc) {
		maxDistJacc = dJacc;
		maxJaccInd.clear();
		maxJaccInd.add(iI);
	    }
	    if( (dJaro == maxDistJaro) && (dJaro > 0.0)) {
		maxJaroInd.add(iI);
	    }
	    if(dJaro > maxDistJaro) {
		maxDistJaro = dJaro;
		maxJaroInd.clear();
		maxJaroInd.add(iI);
	    }
	    if( (dJen2 == maxDistJen2) && (dJen2 > 0.0)) {
		maxJShan2Ind.add(iI);
	    }
	    if(dJen2 > maxDistJen2) {
		maxDistJen2 = dJen2;
		maxJShan2Ind.clear();
		maxJShan2Ind.add(iI);
	    }
	    
	    if( (dL == maxDistL) && (dL >= 0.0)) {
		maxLevInd.add(iI);
	    }
	    if(dL > maxDistL) {
		maxDistL = dL;
		maxLevInd.clear();
		maxLevInd.add(iI);
	    }
	}

	//now, print out the feature function values for classifying
	for(int i = 0; i < refStrings.size(); i++) {
	    Integer iI = new Integer(i);
	    vals[i] = "";
	    int score = 0;
	    //max jen shan feature
	    if(maxJShanInd.contains(iI)) {
		vals[i] += "1|";
	    }
	    else {
		vals[i] += "0|";
	    }
	    if(maxSmithInd.contains(iI)) {
		vals[i] += "1|";
	    }
	    else {
		vals[i] += "0|";
	    }
	    if(maxSoundexInd.contains(iI)) {
		vals[i] += "1|";
	    }
	    else {
		vals[i] += "0|";
	    }
	    
	    if(maxStemInd.contains(iI)) {
		vals[i] += "1|";
	    }
	    else {
		vals[i] += "0|";
	    }
	    if(maxJaroInd.contains(iI)) {
		vals[i] += "1|";
	    }
	    else {
		vals[i] += "0|";
	    }
	    if(maxJaccInd.contains(iI)) {
		vals[i] += "1|";
	    }
	    else {
		vals[i] += "0|";
	    }
	    if(maxJShan2Ind.contains(iI)) {
		vals[i] += "1|";
	    }
	    else {
		vals[i] += "0|";	       
	    }
	    if(maxLevInd.contains(iI)) {
		vals[i] += "1|";
	    }
	    else {
		vals[i] += "0|";
	    }

	}
	return vals;
    }

}
