package edu.isi.integration.WebPhoebus;

import edu.isi.integration.BSL.*; 
import edu.isi.integration.Phoebus.*;
import edu.isi.integration.Commons.*;

import java.util.*;
import java.sql.*;

public class WebPhoebus {
    /*
      This method is for using Phoebus with a servlet, for example. It can read in unstructured data, and assuming a model is already trained, it will structure it. This is the application of Phoebus. It assumes you have a blocking scheme already, and then you can run it and get candidates, etc.
     */
    public String structureData(String refSetXMLFile, String refSetName[], String postSetXMLFile, 
			      String phoebusConfigXMLFile,String commonAttributeXMLFile, String chosenCommonScores[]) {
	String outputXML = "";

	//set up to run for multiple reference sets
	DataSet refSetArr[] = new DataSet[refSetName.length];
	DataSet matchTables[] = new DataSet[refSetName.length];
	//since we may be doing different attributes for the post depending on which
	//reference set we are using, so the below dataset holds all of them and their tables...
	DataSet finalPostSet = new DataSet();

	XMLReader xr = new XMLReader();

	TableMaker tm = new TableMaker(); //used to make tables later on...

	for(int currRefSet = 0; currRefSet < refSetName.length; currRefSet++) {
	    DataSet postSet = new DataSet();
	    String currRefSetName = refSetName[currRefSet];
	    //create an instance
	    refSetArr[currRefSet] = new DataSet();
	    xr.retreiveReferenceSet(refSetXMLFile,currRefSetName,refSetArr[currRefSet]);
	    refSetArr[currRefSet].setTableVals();
	    
	    HashMap commonScores = new HashMap();
	    if(currRefSet == 0) { //for now, we only do the common scores stuff once, should only need to..
		commonScores = xr.retrieveCommonScores(commonAttributeXMLFile, chosenCommonScores);
	    }
	    //System.out.println("CS: "+commonScores);

	    //set up generic temporary table...
	    //First, read in where we can make/put this table...
	    postSet.setName("posts");
	    xr.retrieveDatabaseInformation(phoebusConfigXMLFile, postSet);
	    postSet.addAllAttributes(refSetArr[currRefSet].getAttributes());
	    postSet.addAllAttributes(new Vector((Collection)commonScores.keySet()));
	    postSet.addAttribute("post");
	    //Make the temporary table using the information we have so far
	    tm.createTable(postSet);//we pass the dataset object so we can get all of the data for it like url, login, ...

	    //now we have to fill up this table with ALL of the correct attributes
	    if(currRefSet == 0) {
		finalPostSet.setName("finalPostSet");
		finalPostSet.addAttribute("post");
		finalPostSet.addAllAttributes(new Vector((Collection)commonScores.keySet()));
		for(int x = 0; x < refSetName.length; x++) {
		    finalPostSet.addAllAttributes(refSetArr[x].getAttributes());
		}
		finalPostSet.setURL(postSet.getURL());
		finalPostSet.setLogin(postSet.getLogin());
		finalPostSet.setPass(postSet.getPass());
		finalPostSet.setTableName(postSet.getTableName()+"_FINAL_SET");
		finalPostSet.setTableId(postSet.getTableId());
		tm.createTable(finalPostSet);
	    }

	    //create the matches table too...
	    matchTables[currRefSet] = new DataSet();
	    matchTables[currRefSet].setName("matches");
	    matchTables[currRefSet].setTableName(postSet.getTableName()+"_"+refSetArr[currRefSet].getTableName()+"_MATCHES");
	    matchTables[currRefSet].setURL(postSet.getURL());
	    matchTables[currRefSet].setLogin(postSet.getLogin());
	    matchTables[currRefSet].setPass(postSet.getPass());
	    matchTables[currRefSet].addAttribute(postSet.getTableId());
	    matchTables[currRefSet].addAttribute(refSetArr[currRefSet].getTableId());
	    tm.createTable(matchTables[currRefSet]);

	    if(currRefSet == 0) {
		tm.fillTableWithPosts(postSetXMLFile, postSet, 
				      matchTables[currRefSet].getTableName(), refSetArr[currRefSet].getTableId());//fill up the table with the XML data
	    }

	    //update the attributes to be nice blank strings, instead of nulls
	    Vector refSchema = refSetArr[currRefSet].getAttributes();
	    for(int zz = 0; zz < refSchema.size(); zz++) {
		tm.updateField((String)refSchema.get(zz),"",postSet);
	    }
	    if(currRefSet == 0) {
		Set cScoreSet = commonScores.keySet();
		Vector cScores = new Vector((Collection)cScoreSet);
		for(int aa = 0; aa < cScores.size(); aa++) {
		    tm.updateField((String)cScores.get(aa),"",postSet);
		}
	    }
	
	    postSet.setTableVals();
	    if(currRefSet == 0) {
		finalPostSet.insertAttributesFromOtherSet(postSet);
	    }
	    //System.out.println(postSet);

	    //now, we need to set the data set up for blocking (this means copying the post to all attributes 
	    //we are going to use for blocking...
	    Vector blockingScheme = new Vector();
	    xr.retreiveBlockingScheme(refSetXMLFile,currRefSetName,blockingScheme);

	    //this is the path to the svm executables and the temporary files that we create...
	    String dirPath = xr.retrieveFilePath(phoebusConfigXMLFile);

	    //now get the candidates using the scheme...
	    BSL bsl = new BSL();
	    bsl.setPath(dirPath);
	    //set up for blocking...(copy post to all attributes)
	    bsl.setUpTable(postSet, blockingScheme, "blocking");
	    bsl.runBlockingScheme(blockingScheme, postSet, refSetArr[currRefSet], true);
	    //now, clear out the attributes in the data set (remove the copy of the post from the attributes...)
	    bsl.setUpTable(postSet, blockingScheme, "phoebus");

	    Phoebus p = new Phoebus();
	    p.setMode("test");
	    p.setPath(dirPath);
	    p.setCandidateFile(postSet.getTableName()+"_finalCands.txt");
	    p.setClassFile(postSet.getTableName()+"Class.dat");
	    p.buildSVMClassData(postSet, refSetArr[currRefSet], matchTables[currRefSet]);
	    p.runSVMClassifier(postSet.getTableName()+"Class.dat",refSetArr[currRefSet].getTableName()+"ClassModel",postSet.getTableName()+"ClassPreds.txt");
	    p.getMatches(postSet.getTableName()+"Class.dat",postSet.getTableName()+"ClassPreds.txt",refSetArr[currRefSet],postSet,matchTables[currRefSet]);
	    p.setSegFile(postSet.getTableName()+"TestSeg.dat");
	    p.buildSVMSegmentationData(postSet, refSetArr[currRefSet], matchTables[currRefSet], commonScores);
	    p.runSVMSegmenter(postSet.getTableName()+"TestSeg.dat",refSetArr[currRefSet].getTableName()+"SegModel",postSet.getTableName()+"SegPreds.txt");
	    p.getLabeledTokens(postSet.getTableName()+"TestSeg.dat",postSet.getTableName()+"SegPreds.txt",postSet,commonScores);    
	    p.postProcessFields(postSet, refSetArr[currRefSet], matchTables[currRefSet]);
	    //matchTables[currRefSet].closeConnection();
	    //remove all of the files we created...
	    p.removeTemporaryFile(postSet.getTableName()+"_finalCands.txt");
	    p.removeTemporaryFile(postSet.getTableName()+"Class.dat");
	    p.removeTemporaryFile(postSet.getTableName()+"ClassPreds.txt");
	    p.removeTemporaryFile(postSet.getTableName()+"TestSeg.dat");
	    p.removeTemporaryFile(postSet.getTableName()+"SegPreds.txt");

	    //now that we are done we can clean up...
	    finalPostSet.updateAttributesFromOtherSet(postSet);
	    //tm.cleanUp(postSet);
	}
	for(int a = 0; a < refSetArr.length; a++) {
	    matchTables[a].openConnection();
	    matchTables[a].closeConnection();
	}
	XMLWriter xmw = new XMLWriter();
	outputXML = xmw.writeXMLOutputOfPosts(refSetArr, finalPostSet, matchTables, postSetXMLFile);
	//clean up the matches table too...
	for(int a = 0; a < refSetArr.length; a++) {
	    //matchTables[a].closeConnection();
	    //tm.cleanUp(matchTables[a]);
	}
	//tm.cleanUp(finalPostSet);
	return outputXML;
    }

}